import functools
import inspect
import typing
import yaml

from horkos import checks as hs_checks
from horkos import errors
from horkos import _fields
from horkos import _schemaomatic
from horkos import types

Schemaable = typing.Union[str, dict, _schemaomatic.Schema, typing.TextIO]


def _schema_from_dict(
        raw_schema: dict,
        custom_checkers: typing.Dict[str, typing.Callable] = None
) -> _schemaomatic.Schema:
    """
    Create a schema from a raw schema dictionary.

    :param raw_schema:
        A raw schema matching the expected json/yaml schema format.
    :param custom_checkers:
        A dictionary of custom checkers. This dictionary should map the name
        of the check to a function that generates the check function
    :return:
        The resulting schema.
    """
    standardized = _standardize_schema(raw_schema)
    check_map = hs_checks.CHECK_MAP.copy()
    check_map.update(custom_checkers or {})
    fields = []
    raw_fields = standardized['fields']
    for field_name, field_body in raw_fields.items():
        field_checks = []
        for i, raw_check in enumerate(field_body['checks']):
            position = f'fields.{field_name}.checks[{i}]'
            check_name = raw_check['name']
            if check_name not in check_map:
                raise errors.SchemaValidationError(
                    f'In "{position}" check name "{check_name}" is unknown.'
                )
            check = _create_check_func(
                check_map[check_name],
                raw_check['args'],
                position
            )
            field_checks.append(check)
        fields.append(_fields.Field(
            field_name,
            types.TYPE_MAP[field_body['type']],
            field_body['description'],
            field_body['required'],
            field_body['nullable'],
            field_checks,
            field_body['labels'],
        ))
    return _schemaomatic.Schema(
        standardized['name'],
        standardized['description'],
        standardized['labels'],
        fields,
    )

def _load_yaml(
        filepath_or_handle: typing.Union[str, typing.TextIO]
) -> dict:
    """
    Read the given yaml file.

    :param filepath_or_handle:
        The path to the yaml file to convert into a schema.
    :return:
        The resulting schema.
    """
    if not isinstance(filepath_or_handle, str):
        return yaml.safe_load(filepath_or_handle)
    with open(filepath_or_handle) as f:
        return yaml.safe_load(f)


def _create_check_func(
        check: typing.Callable, args: dict, position: str
) -> typing.Callable:
    """
    Create the actual check function, baking the args into it. The returned
    function should only expect a single positional argument to check.

    :param check:
        The base check. This is either a class that will be initialized
        with the given arguments, or a function that will be partially filled
        with the given args.
    :param args:
        The arguments to initialize the check with.
    :param position:
        The position of the check within the schema. This is used to provide
        more clear error messages.
    :return:
        A callable that expects a single positional argument that it will
        check.
    """
    try:
        if inspect.isclass(check):
            return check(**args)
        signature = inspect.signature(check)
        signature.bind('value', **args)
    except TypeError:
        raise errors.SchemaValidationError(
            f'Invalid arguments in "{position}"'
        )
    return functools.partial(check, **args)


def _standardize_schema(raw_schema: dict) -> dict:
    """
    Standardize the given raw_schema with all optional fields filled out
    with defaults. If the raw schema is not valid a SchemaValidationError will
    be raised.

    :param raw_schema:
        The schema to standardize.
    :return:
        The standardized schema.
    """
    standardized = {}
    top = [
        ('name', str, True),
        ('description', str, False),
        ('labels', dict, False),
        ('fields', dict, False),
    ]
    for key, type_, required in top:
        if key not in raw_schema and required:
            raise errors.SchemaValidationError(f'"{key}" is required')
        value = raw_schema[key] if key in raw_schema else type_()
        if not isinstance(value, type_):
            raise errors.SchemaValidationError(
                f'"{key}" must be of type {type_.__name__}'
            )
        standardized[key] = value
    standardized['fields'] = {
        field_id: _standardize_field(raw_field, field_id)
        for field_id, raw_field in raw_schema['fields'].items()
    }
    return standardized


def _standardize_field(raw_field: dict, field_id: str) -> dict:
    """
    Standardize a given raw field with all the optional parts filled out
    with defaults. If the raw field can't be validated a SchemaValidationError
    will be raised.

    :param raw_field:
        The raw field to standardize.
    :param field_id:
        The id of the field being standardized.
    :return:
        The standardized field.
    """
    standardized = {}
    top = [
        ('type', str, True, None),
        ('description', str, False, str),
        ('labels', dict, False, dict),
        ('checks', list, False, list),
        ('required', bool, False, lambda: True),
        ('nullable', bool, False, bool),
    ]
    for key, type_, required, default in top:
        working = f'fields.{field_id}.{key}'
        if key not in raw_field and required:
            raise errors.SchemaValidationError(f'"{working}" is required')
        value = raw_field[key] if key in raw_field else default()
        if not isinstance(value, type_):
            raise errors.SchemaValidationError(
                f'"{working}" must be of type {type_.__name__}'
            )
        standardized[key] = value
    standardized['checks'] = [
        _standardize_check(raw_check, f'fields.{field_id}.check[{i}]')
        for i, raw_check in enumerate(raw_field.get('checks') or [])
    ]
    return standardized


def _standardize_check(
        raw_check: typing.Union[str, dict], prefix: str
) -> dict:
    """
    Convert a check into the standard format. If the check cannot be
    standardize a schema validation error is raised.

    :param raw_check:
        The raw_check to standardize.
    :param prefix:
        The document position prefix to integrate into error messages.
    :return:
        A standard check dictionary.
    """
    if isinstance(raw_check, str):
        return {'name': raw_check, 'args': {}}
    if isinstance(raw_check, dict):
        if 'name' not in raw_check:
            raise errors.SchemaValidationError(
                f'"{prefix}.name" is required'
            )
        if not isinstance(raw_check['name'], str):
            raise errors.SchemaValidationError(
                f'"{prefix}.name" must be a string'
            )
        if 'args' in raw_check and not isinstance(raw_check['args'], dict):
            raise errors.SchemaValidationError(
                f'"{prefix}.args" must be a dictionary'
            )
        return {'name': raw_check['name'], 'args': raw_check.get('args', {})}
    raise errors.SchemaValidationError(
        f'"{prefix}" must be a string or dictionary'
    )


def load_schema(
        schema: Schemaable,
        custom_checkers: typing.Dict[str, typing.Callable] = None
) -> _schemaomatic.Schema:
    """
    Load a schema from a file or existing schema object.

    :param schema:
        The schema to load. If a string is given, it is assumed to be a path
        to a `.json` or `.yaml` file defining the schema, alternatively a file
        handle can be passed directly. If a dictionary is given it will be
        assumed to be equivalent to the contents of a schema file.
    :param custom_checkers:
        A dictionary of custom checkers. This dictionary should map the name
        of the check to a function that generates the check function. This
        can be used to extend checks beyond those defined within `horkos`.
    :return:
        A Schema object.
    """
    working = schema
    custom_checkers = custom_checkers or {}
    if isinstance(working, _schemaomatic.Schema):
        return working
    if not isinstance(working, dict):
        working = _load_yaml(working)
    return _schema_from_dict(
        working, custom_checkers=custom_checkers
    )



from unittest import mock

import pytest

import horkos
from horkos.critiquer import _casing


def test_schema_uniform_field_casing_check_with_critique():
    """Should identify mixed casing conventions within schema"""
    schema = horkos.load_schema({
        'name': 'ds1',
        'fields': {
            'fooBar': {'type': 'integer'}, 'fiz_buzz': {'type': 'integer'}
        }
    })

    result = _casing.schema_uniform_field_casing_check(schema)

    assert result
    expected = 'camelCase and snake_case are most common with 1 matching.'
    assert expected in result[0].message


def test_schema_uniform_field_casing_check_without_critique():
    """Should return none if the casing is consistent"""
    schema = horkos.load_schema({
        'name': 'ds1',
        'fields': {
            'foobar': {'type': 'integer'}, 'fiz_buzz': {'type': 'integer'}
        }
    })

    result = _casing.schema_uniform_field_casing_check(schema)

    assert not result


def test_relative_field_casing_check_with_critique():
    """Should identify mixed field conventions between schema and catalog."""
    field = {'type': 'integer'}
    foo_schema = horkos.load_schema({
        'name': 'foo',
        'fields': {'fooBar': field, 'fiz_buzz': field}
    })
    baz_schema = horkos.load_schema({
        'name': 'baz',
        'fields': {'fooBar': field, 'boo_far': field, 'biz_fuzz': field}
    })
    catalog = horkos.Catalog([foo_schema, baz_schema])
    schema = horkos.load_schema({'name': 'baz', 'fields': {'bizFuzz': field}})

    result = _casing.relative_field_casing_check(schema, catalog)

    assert result
    expected = (
        'The casing of bizFuzz does not match the dominant patterns of the '
        'catalog'
    )
    assert expected in result[0].message


def test_relative_field_casing_check_with_schema_in_catalog():
    """If the schema is already in the catalog an error should be raised."""
    catalog = mock.MagicMock(schemas={'foo': 'bar'})
    schema = mock.MagicMock()
    schema.name = 'foo'

    with pytest.raises(ValueError):
        _casing.relative_field_casing_check(schema, catalog)


def test_relative_field_casing_without_critique():
    """Should return None if their are no issues."""
    foobar_field = mock.MagicMock()
    foobar_field.name = 'fooBar'
    fizbuzz_field = mock.MagicMock()
    fizbuzz_field.name = 'fizBuzz'
    boofar_field = mock.MagicMock()
    boofar_field.name = 'booFar'
    bizfuzz_field = mock.MagicMock()
    bizfuzz_field.name = 'bizFuzz'
    catalog = mock.MagicMock(
        schemas=[
            mock.MagicMock(fields=[foobar_field, fizbuzz_field]),
            mock.MagicMock(fields=[foobar_field, boofar_field, bizfuzz_field]),
        ]
    )
    schema = mock.MagicMock(fields=[foobar_field, bizfuzz_field])

    result = _casing.relative_field_casing_check(schema, catalog)

    assert not result


def test_catalog_field_casing_check_with_critique():
    """Should be able to identify inconsistent naming in a catalog of data."""
    field = {'type': 'integer'}
    schemas = [
        horkos.load_schema({
            'name': 'foo',
            'fields': {'fooBar': field, 'fiz_buzz': field}
        }),
        horkos.load_schema({
            'name': 'baz',
            'fields': {'fooBar': field, 'boo_far': field, 'biz_fuzz': field}
        }),
    ]
    catalog = horkos.Catalog(schemas)

    result = _casing.catalog_field_casing_check(catalog)

    assert result
    expected = (
        'No clear casing convention. snake_case is most '
        'common with 60% matching.'
    )
    assert expected in result[0].message


def test_catalog_field_casing_check_without_critique():
    """Should return None if all fields are consistent."""
    field = {'type': 'integer'}
    schemas = [
        horkos.load_schema({
            'name': 'foo',
            'fields': {'foobar': field, 'fiz_buzz': field}
        }),
        horkos.load_schema({
            'name': 'baz',
            'fields': {'foobar': field, 'boo_far': field, 'biz_fuzz': field}
        }),
    ]
    catalog = horkos.Catalog(schemas)

    result = _casing.catalog_field_casing_check(catalog)

    assert not result


def test_schema_prefer_snake_case_with_critique():
    """
    Should be able to identify when snake_case isn't used in a schema
    and recommend its usage.
    """
    field = {'type': 'integer'}
    schema = horkos.load_schema({
        'name': 'foo',
        'fields': {'fooBar': field, 'fizBuzz': field}
    })

    result = _casing.schema_prefer_snake_case(schema)

    assert result
    expected = 'snake_case field names are recommended'
    assert expected in result[0].message


def test_schema_prefer_snake_case_without_critique():
    """If snake_case is already used, then return None."""
    field = {'type': 'integer'}
    schema = horkos.load_schema({
        'name': 'foo',
        'fields': {'foobar': field, 'fiz_buzz': field}
    })

    result = _casing.schema_prefer_snake_case(schema)

    assert not result


def test_catalog_prefer_snake_case_with_critique():
    """
    Should be able to identify when snake_case isn't used in a schema
    and recommend its usage.
    """
    field = {'type': 'integer'}
    schemas = [
        horkos.load_schema({
            'name': 'foo',
            'fields': {'Foobar': field, 'fiz_buzz': field}
        }),
        horkos.load_schema({
            'name': 'baz',
            'fields': {'Foobar': field, 'BooFar': field, 'biz_fuzz': field}
        }),
    ]
    catalog = horkos.Catalog(schemas)

    result = _casing.catalog_prefer_snake_case(catalog)

    assert result
    expected = 'snake_case field names are recommended'
    assert expected in result[0].message


def test_catalog_prefer_snake_case_without_critique():
    """If snake_case is already used, then return None."""
    field = {'type': 'integer'}
    schemas = [
        horkos.load_schema({
            'name': 'foo',
            'fields': {'foobar': field, 'fiz_buzz': field}
        }),
        horkos.load_schema({
            'name': 'baz',
            'fields': {'foobar': field, 'boo_far': field}
        }),
    ]
    catalog = horkos.Catalog(schemas)

    result = _casing.catalog_prefer_snake_case(catalog)

    assert not result


def test_most_common_casing_type_when_empty():
    """If empty an empty list should be returned."""
    values = ['foo_Bar-baz']

    result = _casing._most_common_casing_type(values)

    assert not result[0]
    assert result[1] == 0



import typing

from horkos import types


class Field:
    """
    The definition of a field within a schema.

    :vartype name: str
    :ivar name:
        The name of the field.
    :vartype type: types.FieldType
    :ivar type:
        The type of the field.
    :vartype description: str
    :ivar description:
        A detailed description of the field.
    :vartype required: bool
    :ivar required:
        Whether the field is required to be present.
    :vartype nullable: bool
    :ivar nullable:
        Whether the field should accept null values.
    :vartype checks: typing.List[typing.Callable]
    :ivar checks:
        A list of callables accepting a single value to validate.
    :vartype labels: dict
    :ivar labels:
        A space for unstructured information regarding the field.
    """

    def __init__(
            self,
            name: str,
            type_: types.FieldType,
            description: str = None,
            required: bool = True,
            nullable: bool = False,
            checks: typing.List[typing.Callable] = None,
            labels: dict = None,
    ):
        """Constructor for field."""
        self.name = name
        self.type = type_
        self.description = description
        self.required = required
        self.nullable = nullable
        self.checks = checks or []
        self.labels = labels or []



# pylint: disable=broad-except
import functools
import typing

from horkos import errors


def _allow_nulls(func: typing.Callable) -> typing.Callable:
    """Create a function that immediately returns None if given None."""

    @functools.wraps(func)
    def wrapper(value: typing.Any) -> typing.Any:
        """Return None if None is given."""
        if value is None:
            return value
        return func(value)

    return wrapper


def _all_errors_to_casting_error(func: typing.Callable) -> typing.Callable:
    """Create a function that converts exceptions to casting errors."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        """Convert all exceptions to CastingErrors."""
        try:
            return func(*args, **kwargs)
        except errors.CastingError as err:
            raise err
        except Exception:
            raise errors.CastingError('cannot cast')

    return wrapper


@_all_errors_to_casting_error
@_allow_nulls
def _integer(value: typing.Any) -> int:
    """Convert the given value into an integer."""
    if isinstance(value, float) and (int(value) - value) != 0:
        raise errors.CastingError('cannot losslessly cast')
    return int(value)


@_all_errors_to_casting_error
@_allow_nulls
def _bool(value: typing.Any) -> bool:
    """Convert the given value to a boolean"""
    if isinstance(value, bool):
        return value
    if isinstance(value, str) and value.lower() == 'true' or value == '1':
        return True
    if isinstance(value, str) and value.lower() == 'false' or value == '0':
        return False
    if isinstance(value, int) and value == 0:
        return False
    if isinstance(value, int) and value == 1:
        return True
    raise errors.CastingError('cannot unambiguously cast')


@_all_errors_to_casting_error
@_allow_nulls
def _float(value: typing.Any) -> float:
    """Convert the given value to a float"""
    return float(value)


@_all_errors_to_casting_error
@_allow_nulls
def _string(value: typing.Any) -> str:
    """Convert the given value to a string."""
    if isinstance(value, (str, int, float)):
        return str(value)
    raise errors.CastingError('cannot unambiguously cast')


class FieldType:
    """The base type for all field types."""

    name: str
    _caster: typing.Callable

    @classmethod
    def cast(cls, value: typing.Any):
        """Cast the incoming value to the desired type."""
        return cls._caster(value)


class Boolean(FieldType):
    """
    A value that is either true or false. :code:`boolean` fields will cast
    strings that case insensitively match :code:`"true"` or :code:`"false"`
    exactly to :code:`True` and :code:`False` respectively. It will cast
    integers :code:`0` and :code:`1` to :code:`False` and :code:`True`
    respectively.
    """
    name = 'boolean'
    _caster = _bool


class Float(FieldType):
    """
    A numerical value with decimal components. :code:`float` fields will cast
    integers, as well as float strings and integer strings to float values.
    """
    name = 'float'
    _caster = _float


class Integer(FieldType):
    """
    A whole number value. :code:`integer` fields will cast integer strings
    to integers, they will also cast float values to integers if there is no
    decimal component. Float values with decimal components will fail to
    cast to integers.
    """
    name = 'integer'
    _caster = _integer


class String(FieldType):
    """
    Any string value. :code:`horkos` will only cast values that can be cast
    unambiguously to a string such as integers and floats. Other values
    that have a string encoding in python will not be cast to a string if
    there isn't exactly one reasonable string encoding. Examples of this are
    booleans (since :code:`True`, :code:`TRUE`, and :code:`true` are all
    equally reasonable) or dictionaries (since :code:`{'foo': 'bar'}` and
    :code:`{"foo": "bar"}` are equally reasonable).
    """
    name = 'string'
    _caster = _string


ALL_TYPES = [Boolean, Float, Integer, String]
TYPE_MAP = {t.name: t for t in ALL_TYPES}



import typing

import pytest

from horkos import types
from horkos import errors


def test_str_less_aggressive():
    """The string caster in _casters should not aggressively cast to strings."""
    value = {'foo': 'bar'}

    with pytest.raises(errors.CastingError) as err:
        types._string(value)

    assert 'cannot unambiguously cast' in str(err.value)


SCENARIOS = [
    ('true', True),
    ('False', False),
    (0, False),
    (1, True),
    (True, True),
]

@pytest.mark.parametrize('value,expected', SCENARIOS)
def test_boolean(value: typing.Union[bool, str, int], expected: bool):
    """Should be able to cast value to booleans."""
    result = types._bool(value)

    assert result == expected



import typing


class CheckMetadata(typing.NamedTuple):
    """The name and args associated with a check."""

    name: str
    args: dict


def get_check_metadata(check: typing.Callable) -> CheckMetadata:
    """Get the metadata associated with a given check."""
    name = getattr(check, 'name', None) or getattr(check, '__name__')
    args = getattr(check, 'args', {})
    return CheckMetadata(name, args)



import collections
import typing

from horkos import errors
from horkos import inspector
from horkos import _fields


class Schema:
    """
    A formal specification of a dataset. It defines all fields (or columns) a
    dataset is expected to have as well as the properties of each field.

    :vartype name: str
    :ivar name:
        The name of the dataset defined by the schema.
    :vartype name: str
    :ivar description:
        A detailed description of the dataset defined by the schema.
    :vartype labels: dict
    :ivar labels:
        A dictionary of key value pairs containing organization
        specific structured information about the dataset. These values have
        no function within `horkos`.
    :vartype fields: typing.List[horkos.Field]
    :ivar fields:
        A list of fields expected within the dataset.
    :param name:
        The name of the dataset described by the schema. This is used to
        provide more descriptive error messages and to provide an identity
        while being used as part of a Catalog.
    :param description:
        A detailed description of the dataset defined by the schema.
    :param labels:
        A dictionary of minimally structured key value pairs intended for
        storing organization specific information relating to the dataset.
    :param fields:
        A list of fields (or columns) that are expected within the dataset.
        Each field must have its own unique name.
    """

    def __init__(
            self,
            name: str = None,
            description: str = None,
            labels: dict = None,
            fields: typing.List[_fields.Field] = None,
    ):
        """Class constructor."""
        self.name = name
        self.description = description
        self.labels = dict(labels or {})
        fields = fields or []
        common = collections.Counter(f.name for f in fields).most_common(1)
        if common and common[0][1] > 1:
            msg = (
                f'Field named "{common[0][0]}" appears {common[0][1]} times.'
                ' Field names should be unique within a schema.'
            )
            raise ValueError(msg)
        self.fields = list(fields)

    def _cast(self, record: dict) -> dict:
        """
        Cast all the values contained within the record to the types declared
        within the schema, removing all undeclared fields from the record.

        :param record:
            A dictionary mapping field names to field values.
        :return:
            The record with values cast to the declared types.
        """
        error_list = []
        cast = {}
        for field in self.fields:
            try:
                value = record.get(field.name)
                cast[field.name] = field.type.cast(value)
            except errors.CastingError as err:
                error_list.append(
                    f'{str(err)} {value} to {field.type.name} in {field.name}'
                )
        if error_list:
            msg = f'Casting failed: {", ".join(error_list)}'
            raise errors.RecordValidationError(msg)
        return cast

    def _apply_checks(self, record: dict):
        """
        Apply field checks to the incoming record, raising an error if
        any of them fail.
        """
        error_list = []
        for field in self.fields:
            value = record[field.name]
            if value is None:
                continue
            for check in field.checks:
                if not check(value):
                    expanded = inspector.get_check_metadata(check)
                    error_list.append(
                        f'value of "{value}" for {field.name} did not pass '
                        f'{expanded.name} check'
                    )
        if error_list:
            msg = f'Checks failed: {", ".join(error_list)}'
            raise errors.RecordValidationError(msg)

    def process(self, record: dict) -> dict:
        """
        Process a record against the schema. This process includes:

            1. Confirming all required fields are present.
            2. Casting all field values to their expected type.
            3. Confirming there are no null values in non-nullable fields.
            4. Confirm all non-null values pass their field's checks.

        The cast and validated record is returned.

        :param record:
            A dictionary mapping field names to field values.
        :return:
            The processed record with values cast and validated.
        """
        missing_fields = [
            f.name for f in self.fields if f.required and f.name not in record
        ]
        if missing_fields:
            msg = f'Required fields are missing: {", ".join(missing_fields)}'
            raise errors.RecordValidationError(msg)

        cast = self._cast(record)

        null_fields = [
            f.name
            for f in self.fields if not f.nullable and cast[f.name] is None
        ]
        if null_fields:
            msg = f'Non-nullable fields are null: {", ".join(null_fields)}'
            raise errors.RecordValidationError(msg)

        self._apply_checks(cast)
        return cast



from horkos._cataloger import Catalog
from horkos._fields import Field
from horkos._schemaomatic import Schema
from horkos._yaml_parser import load_schema


__all__ = ('Catalog', 'Field', 'load_schema', 'Schema')



import os
import typing

import pytest

import horkos
from horkos import errors
from horkos import _schemaomatic
from horkos import types


def test_process_raises_all_casting_errors():
    """Should raise all errors when present."""
    fields = [
        horkos.Field('bool_field', types.Boolean),
        horkos.Field('float_field', types.Float),
        horkos.Field('int_field_1', types.Integer),
        horkos.Field('int_field_2', types.Integer),
        horkos.Field('str_field', types.String),
    ]
    schema = _schemaomatic.Schema(fields=fields)
    record = {
        'bool_field': 'T',
        'float_field': 'word',
        'int_field_1': 1.123,
        'int_field_2': 'word',
        'str_field': None,
    }

    with pytest.raises(errors.RecordValidationError) as err:
        schema.process(record)

    msg = 'cannot unambiguously cast T to boolean in bool_field'
    assert msg in str(err.value)
    msg = 'cannot cast word to float in float_field'
    assert msg in str(err.value)
    msg = 'cannot losslessly cast 1.123 to integer in int_field_1'
    assert msg in str(err.value)
    msg = 'cannot cast word to integer in int_field_2'
    assert msg in str(err.value)


def test_schema_with_repeated_fields():
    """Repeated fields of the same name should raise an error."""
    field = horkos.Field('field_name', types.Boolean)

    with pytest.raises(ValueError):
        _schemaomatic.Schema(fields=[field, field])



"""
Check that an incoming stream or file of data matches the declared
schema.
"""
import argparse
import csv
import io
import json
import sys
import typing

from horkos import errors
from horkos import _yaml_parser


def configure_parser(parser: argparse.ArgumentParser):
    """Configure the given parser to work for the check command."""
    parser.add_argument(
        '--schema', '-s',
        dest='schema_file',
        required=True,
        type=argparse.FileType('r'),
        help='The schema file describing the data.'
    )
    parser.add_argument(
        'datafile',
        nargs='?',
        type=argparse.FileType('r'),
        default=sys.stdin,
    )


def _convert_to_records(contents: str) -> typing.List[dict]:
    """
    Convert a file's contents into a set of records. The contents are expected
    to be either a csv with a header of column names, a json list, or a
    set of json records seperated by a new line.

    :param contents:
        The contents of the file to parse.
    :return:
        A list of records from the file.
    """
    try:
        rows = json.loads(contents)
        if isinstance(rows, list):
            return rows
    except ValueError:
        pass
    try:
        rows = [json.loads(r) for r in contents.strip().split('\n')]
        if all(isinstance(r, dict) for r in rows):
            return rows
    except ValueError:
        pass
    reader = csv.DictReader(io.StringIO(contents))
    return [
        {k: v if v != '' else None for k, v in r.items()}
        for r in reader
    ]


def main(
        datafile: io.TextIOWrapper,
        schema_file: io.TextIOWrapper,
        **_,
):
    """Validate that the input file matches the given schema."""
    schema = _yaml_parser.load_schema(schema_file)
    contents = datafile.read()
    records = _convert_to_records(contents)
    error_count = 0
    for i, record in enumerate(records):
        try:
            schema.process(record)
        except errors.RecordValidationError as err:
            error_count += 1
            print(f'[Row {i}]: {err}', file=sys.stdout)
    msg = f'{error_count} error{"s" if error_count != 1 else ""} found'
    print(msg)



from unittest import mock

import horkos
from horkos.critiquer import _typing


def test_catalog_type_consistency_check_with_critique():
    """The types of each field should be consistent across schemas."""
    schema_1 = horkos.load_schema({
        'name': 'ds1', 'fields': {'foobar': {'type': 'integer'}}
    })
    schema_2 = horkos.load_schema({
        'name': 'ds2', 'fields': {'foobar': {'type': 'string'}}
    })
    catalog = horkos.Catalog([schema_1, schema_2])

    result = _typing.catalog_type_consistency_check(catalog)

    assert result
    expected = (
        'Catalog describes foobar as integer and string. The type of foobar '
        'should be consistent between schemas.'
    )
    assert expected in result[0].message


def test_catalog_type_consistency_check_without_critique():
    """If types are consistent then no critique should be returned."""
    schema_1 = horkos.load_schema({
        'name': 'ds1', 'fields': {'foobar': {'type': 'integer'}}
    })
    schema_2 = horkos.load_schema({
        'name': 'ds2', 'fields': {'foobar': {'type': 'integer'}}
    })
    catalog = horkos.Catalog([schema_1, schema_2])

    result = _typing.catalog_type_consistency_check(catalog)

    assert not result


def test_relative_type_consistency_check_with_critique():
    """The types of a new schema should be consistent with existing"""
    schema = horkos.load_schema({
        'name': 'ds1', 'fields': {'foobar': {'type': 'integer'}}
    })
    catalog = horkos.Catalog([schema])
    new_schema = horkos.load_schema({
        'name': 'ds2', 'fields': {'foobar': {'type': 'string'}}
    })

    result = _typing.relative_type_consistency_check(new_schema, catalog)

    assert result
    expected = 'other schemas in the catalog have it declared as integer'
    assert expected in result[0].message


def test_relative_type_consistency_check_without_critique():
    """If types are consistent, then no critiques should be raised."""
    schema = horkos.load_schema({
        'name': 'ds1', 'fields': {'foobar': {'type': 'integer'}}
    })
    catalog = horkos.Catalog([schema])
    new_schema = horkos.load_schema({
        'name': 'ds2', 'fields': {'foobar': {'type': 'integer'}}
    })

    result = _typing.relative_type_consistency_check(new_schema, catalog)

    assert not result



import typing

from horkos import _yaml_parser
from horkos import _schemaomatic


class Catalog:
    """
    A collection of schemas.

    :param schemas:
        The schemas to include in the catalog.
    """

    def __init__(
            self,
            schemas: typing.List[_yaml_parser.Schemaable] = None,
    ):
        """Initialize the catalog of data schemas."""
        schemas = schemas or []
        all_schemas = [
            _yaml_parser.load_schema(s) for s in schemas
        ]
        self._schema_map = {s.name: s for s in all_schemas}
        if None in self._schema_map:
            raise ValueError('All schemas within a catalog must be named.')


    def process(
            self,
            name: str,
            record: dict,
    ) -> dict:
        """
        Process a record against a named schema from the catalog.

        Each field within the record will be cast to the type specified in
        the schema and the resulting value validated against the checks
        defined within the schema. If any of the fields
        cannot be successfully cast or any of the checks fail a
        RecordValidationError exception will be raised.

        :param name:
            The name of the schema to process the record against.
        :param record:
            The record to process against the schema. This should be
            a dictionary mapping field names to field values.
        :return:
            The processed record with values cast and validated.
        """
        schema = self._pull_schema(name)
        return schema.process(record)

    def _pull_schema(self, name: str) -> _schemaomatic.Schema:
        """
        Pull the schema with the given name, if no matching schema exists
        raise a value error indicating as much.
        """
        schema = self._schema_map.get(name)
        if schema is None:
            raise ValueError(f'No schema exists for "{name}"')
        return schema

    def update(self, schema: _yaml_parser.Schemaable):
        """
        Update the catalog with the given schema.

        :param schema:
            The schema to add/update to the catalog.
        """
        to_add = _yaml_parser.load_schema(schema)
        if to_add.name is None:
            raise ValueError('All schemas within a catalog must be named.')
        self._schema_map[to_add.name] = to_add

    @property
    def schemas(self) -> typing.List[_schemaomatic.Schema]:
        """All of the schemas contained within the catalog."""
        return list(self._schema_map.values())



import dataclasses


@dataclasses.dataclass(frozen=True)
class Critique:
    """
    A critique of a dataschema.

    :vartype scope: str
    :ivar scope:
        The scope of the critique. A critique can be scoped as
        `schema`, `relative`, or `catalog` corresponding to
        whether it applies to a single schema, an addition of a schema
        to an existing catalog, or a catalog as a whole.
    :vartype code: str
    :ivar code:
        A short code identifying the critique.
    :vartype message: str
    :ivar message:
        A message describing the critique in detail.
    :vartype schema_name: str
    :ivar schema_name:
        The schema the critique is associated with.
    :vartype field_name: str
    :ivar field_name:
        The field the critique is associated with.
    """

    scope: str
    code: str
    message: str
    schema_name: str = None
    field_name: str = None

    def __str__(self):
        """Convert the critique to a string."""
        parts = [self.schema_name, self.field_name]
        parts = [p for p in parts if p is not None]
        name = '.'.join(parts)
        return f'[{self.code}] {name} - {self.message}'
